#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include "assert.h"
#include "aa.h"
#include "block.h"

#define T Block_T
struct T {
  int positions;
  int sequences;
  Tally_T *columns;
};


T
Block_new (int positions) {
  T block = (T) malloc(sizeof(*block));

  block->positions = positions;
  block->sequences = 0;
  block->columns = (Tally_T *) calloc(positions,sizeof(Tally_T));
  return block;
}

Tally_T
Block_get (T block, int i) {
  return block->columns[i];
}
  
void
Block_put (T block, int i, Tally_T tally) {
  block->columns[i] = tally;
}

void
Block_free (T *block) {
  int pos;
  Tally_T head;
  
  for (pos = 0; pos < (*block)->positions; ++pos) {
    head = Block_get(*block,pos);
    Tally_free(&head);
  }
  free((*block)->columns);
  free(*block);
}
    
int
Block_positions (T block) {
  return block->positions;
}

int
Block_sequences (T block) {
  return block->sequences;
}

T
Block_read_new (FILE *fp, int positions) {
  T block = Block_new(positions);
  int pos, seq = 0;
  char c;
  Tally_T head, found;
  
  while (!feof(fp)) {
    while (!feof(fp) && !isalpha(c = fgetc(fp))) ;
    if (feof(fp)) {
      block->sequences = seq;
      return block;
    } else {
      ungetc(c,fp);
      pos = 0;
      while (!feof(fp) && (pos < block->positions) &&
	     ((c = fgetc(fp)) != '\n')) {
	if (isalpha(c)) {
	  if (AA_acd_index(c) != -1) {
	    head = Block_get(block,pos);
	    if (found = Tally_find(head,c)) {
	      Tally_addseq(found,seq);
	    } else {
	      found = Tally_push(head,c);
	      Block_put(block,pos,found);
	      Tally_addseq(found,seq);
	    }
	  }
	  ++pos;
	}
      }
      if (feof(fp)) {
	printf("Error reading blockfile: EOF in middle of a sequence\n");
	exit(1);
      } else if (c == '\n') {
	printf("Error reading blockfile: Sequence is too short\n");
	exit(1);
      } else {
	while (!feof(fp) && ((c = fgetc(fp)) != '\n')) ;
      }
      ++seq;
    }
  }
  return block;
}

void
Block_print (T block) {
  int pos;

  for (pos = 0; pos < block->positions; ++pos) {
    Tally_print(block->columns[pos]);
  }
}
