#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "intlist.h"
#include "sequence.h"
#include "entry.h"

#define MAXSIGNIF 40
#define MAXPOSITIVES 50000
#define MAXSUBST 100
#define LEFTMARGIN 15
#define RIGHTMARGIN 15

typedef enum{FALSE,TRUE} bool;

static int Num_positives;
static Entry_T Entries[MAXPOSITIVES];


static void
initialize_map (int *map) {
  int i;

  for (i = 0; i < 129; ++i) {
    map[i] = 20;
  }
  map['A'] = 0;  map['B'] = 1;  map['C'] = 2;  map['D'] = 3;
  map['E'] = 4;  map['F'] = 5;  map['G'] = 6;  map['H'] = 7;
  map['I'] = 8;  map['J'] = 20; map['K'] = 9;  map['L'] = 10;
  map['M'] = 11;  map['N'] = 12;  map['O'] = 20; map['P'] = 13;
  map['Q'] = 14;  map['R'] = 15;  map['S'] = 16;  map['T'] = 17;
  map['U'] = 20; map['V'] = 18;  map['W'] = 19;  map['X'] = 20;
  map['Y'] = 21;  map['Z'] = 22;  
  
  map['a'] = 0;  map['b'] = 1;  map['c'] = 2;  map['d'] = 3;
  map['e'] = 4;  map['f'] = 5;  map['g'] = 6;  map['h'] = 7;
  map['i'] = 8;  map['j'] = 20; map['k'] = 9;  map['l'] = 10;
  map['m'] = 11;  map['n'] = 12;  map['o'] = 13; map['p'] = 13; 
  map['q'] = 14;  map['r'] = 15;  map['s'] = 16;  map['t'] = 17;
  map['u'] = 20; map['v'] = 18;  map['w'] = 19;  map['x'] = 20;
  map['y'] = 21;  map['z'] = 22;

  /* map['*'] = 23;  map['-'] = 24; */
}


static int *
map_sequence (char *sequence_char, int length) {
  int *sequence_int;
  int map[129], i;

  sequence_int = (int *) calloc(length,sizeof(int));
  initialize_map(map);
  for (i = 0; i < length; ++i) {
    sequence_int[i] = map[sequence_char[i]];
  }
  return sequence_int;
}


static double
interpolate_score (int score, int *thresholds) {
  double prob0, prob1;
  int sig, sig0, sig1, score0, score1;

  sig = 0;
  while (score <= thresholds[sig]) {
    ++sig;
  }
  if (sig == 0) {
    return pow(10.0,-MAXSIGNIF);
  } else {
    sig1 = MAXSIGNIF - sig;
    sig0 = sig1+1;

    score1 = thresholds[MAXSIGNIF - sig1];
    score0 = thresholds[MAXSIGNIF - sig0];
    prob1 = pow(10.0,-sig1);
    prob0 = pow(10.0,-sig0);
    if (score0 == score1) {
      return prob0;
    } else {
      return ((double) (score - score0)/(double) (score1 - score0)) 
	* (prob1 - prob0) + prob0;
    }
  }
}

static void
open_data_files (FILE **matrices_file, FILE **thresholds_file,
		 FILE **descripts_file, char *data_dir, char *prefix) {
  char *filename, *base;
  int length, baselength = 0;
  
  baselength += strlen(data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    baselength += 1;
  }
  baselength += strlen(prefix);
  base = (char *) calloc(baselength+1,sizeof(char));
  strcpy(base,data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    strcat(base,"/");
  }
  strcat(base,prefix);

  length = strlen(base)+strlen(".mat");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".mat");
  if (!(*matrices_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".thr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".thr");
  if (!(*thresholds_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".des");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".des");
  if (!(*descripts_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);
  free(base);
}

static void
open_data_files_floor (FILE **matrices_file, FILE **thresholds_file,
		       FILE **floors_file, FILE **descripts_file, 
		       char *data_dir, char *prefix) {
  char *filename, *base;
  int length, baselength = 0;
  
  baselength += strlen(data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    baselength += 1;
  }
  baselength += strlen(prefix);
  base = (char *) calloc(baselength+1,sizeof(char));
  strcpy(base,data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    strcat(base,"/");
  }
  strcat(base,prefix);

  length = strlen(base)+strlen(".mat");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".mat");
  if (!(*matrices_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".thr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".thr");
  if (!(*thresholds_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".flr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".flr");
  if (!(*floors_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".des");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".des");
  if (!(*descripts_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);
  free(base);
}

static void
hit_copy (char *segment, char *sequence_char, int start, int width,
	  int length) {
  int l, p;

  for (l = start-LEFTMARGIN, p = 0; l < start; ++l) {
    if (l < 0) {
      segment[p++] = '.';
    } else {
      segment[p++] = sequence_char[l];
    }
  }
  segment[p++] = ' ';
  for ( ; l < start + width; ++l) {
    segment[p++] = sequence_char[l];
  }
  segment[p++] = ' ';
  for ( ; l < start + width + RIGHTMARGIN; ++l) {
    if (l >= length) {
      segment[p++] = '.';
    } else {
      segment[p++] = sequence_char[l];
    }
  }
  segment[p] = '\0';
}


static void
score_sequence (int *sequence_int, char *sequence_char, int length,
		int cutoff_sig, char *prefix, float threshold_info) {
  FILE *mat_fp, *thresh_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[1000];
  double prob;
  short *matrix, *ptr;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh;
  float info;

  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files(&mat_fp,&thresh_fp,&descripts_fp,DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = 0;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + k]];
	}
	if (score > thresh) {
	  if (Num_positives >= MAXPOSITIVES) {
	    printf("Error! Too many positives: Limit is %d\n",
		   MAXPOSITIVES);
	    exit(1);
	  } else {
	    hit_copy(hit,sequence_char,j,width,length);
	    prob = interpolate_score(score,thresholds);
	    fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	    fread(&description[0],sizeof(char),descript_length,descripts_fp);
	    last_descript_ptr = descript_ptr + (long) descript_length;
	    Entries[Num_positives++] = 
	      Entry_new(description,descript_length,j,width,score,prob,info,hit);
	  }
	}
      }
    }
  }
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

static void
score_sequence_floor (int *sequence_int, char *sequence_char, int length,
		      int cutoff_sig, char *prefix, float threshold_info) {
  FILE *mat_fp, *thresh_fp, *floor_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[1000];
  double prob;
  short *matrix, *ptr;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh, *floors;
  float info;

  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files_floor(&mat_fp,&thresh_fp,&floor_fp,&descripts_fp,
			DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));
  floors = (int *) calloc(maxblocksize,sizeof(int));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&floors[0],sizeof(int),width,floor_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    /* j is position along sequence */
    /* k is position along scoring matrix */
    /* i is a matrix pointer */
    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = -thresh;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + k]];
	  if (score < floors[k]) {
	    goto too_low;
	  }
	}
	score += thresh;
	if (Num_positives >= MAXPOSITIVES) {
	  printf("Error! Too many positives: Limit is %d\n",
		 MAXPOSITIVES);
	  exit(1);
	} else {
	  hit_copy(hit,sequence_char,j,width,length);
	  prob = interpolate_score(score,thresholds);
	  fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	  fread(&description[0],sizeof(char),descript_length,descripts_fp);
	  last_descript_ptr = descript_ptr + (long) descript_length;
	  Entries[Num_positives++] = 
	    Entry_new(description,descript_length,j,width,score,prob,info,hit);
	}
      too_low: ;
      }
    }
  }
  free(floors);
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(floor_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

static void
score_sequence_ordered (int *sequence_int, char *sequence_char, int length,
			int cutoff_sig, char *prefix, float threshold_info) {
  FILE *mat_fp, *thresh_fp, *floor_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[1000];
  double prob;
  short *matrix, *perm;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh, *floors;
  float info;

  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files_floor(&mat_fp,&thresh_fp,&floor_fp,&descripts_fp,
			DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));
  floors = (int *) calloc(maxblocksize,sizeof(int));
  perm = (short *) calloc(maxblocksize,sizeof(short));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&perm[0],sizeof(short),width,mat_fp);
    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&floors[0],sizeof(int),width,floor_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    /* j is position along sequence */
    /* k is position along scoring matrix */
    /* i is a matrix pointer */
    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = -thresh;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + perm[k]]];
	  if (score < floors[k]) {
	    goto too_low;
	  }
	}
	score += thresh;
	if (Num_positives >= MAXPOSITIVES) {
	  printf("Error! Too many positives: Limit is %d\n",
		 MAXPOSITIVES);
	  exit(1);
	} else {
	  hit_copy(hit,sequence_char,j,width,length);
	  prob = interpolate_score(score,thresholds);
	  fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	  fread(&description[0],sizeof(char),descript_length,descripts_fp);
	  last_descript_ptr = descript_ptr + (long) descript_length;
	  Entries[Num_positives++] = 
	    Entry_new(description,descript_length,j,width,score,prob,info,hit);
	}
      too_low: ;
      }
    }
  }
  free(perm);
  free(floors);
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(floor_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

static void
score_sequence_bench (int *sequence_int, char *sequence_char, int length,
		      int cutoff_sig, char *prefix, float threshold_info) {
  FILE *mat_fp, *thresh_fp, *floor_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[1000];
  double prob;
  short *matrix, *ptr;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh, *floors;
  float info;
  int potential_positions, actual_positions, skips;

  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files_floor(&mat_fp,&thresh_fp,&floor_fp,&descripts_fp,
			DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));
  floors = (int *) calloc(maxblocksize,sizeof(int));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    potential_positions = 0;
    actual_positions = 0;
    skips = 0;

    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&floors[0],sizeof(int),width,floor_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    /* j is position along sequence */
    /* k is position along scoring matrix */
    /* i is a matrix pointer */
    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = -thresh;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + k]];
	  if (score < floors[k]) {
	    actual_positions += k+1;
	    potential_positions += width;
	    ++skips;
	    goto too_low;
	  }
	}
	score += thresh;
	actual_positions += width;
	potential_positions += width;
	if (Num_positives >= MAXPOSITIVES) {
	  printf("Error! Too many positives: Limit is %d\n",
		 MAXPOSITIVES);
	  exit(1);
	} else {
	  hit_copy(hit,sequence_char,j,width,length);
	  prob = interpolate_score(score,thresholds);
	  fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	  fread(&description[0],sizeof(char),descript_length,descripts_fp);
	  last_descript_ptr = descript_ptr + (long) descript_length;
	  Entries[Num_positives++] = 
	    Entry_new(description,descript_length,j,width,score,prob,info,hit);
	}
      too_low: ;
      }
    }
    printf("%d %f\n",width,
	   1.0 - (double) actual_positions/(double) potential_positions);
  }
  free(floors);
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(floor_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

static void
score_sequence_bench_ordered
(int *sequence_int, char *sequence_char, int length,
 int cutoff_sig, char *prefix, float threshold_info) {
  FILE *mat_fp, *thresh_fp, *floor_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[1000];
  double prob;
  short *matrix, *perm;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh, *floors;
  float info;
  int potential_positions, actual_positions, skips;

  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files_floor(&mat_fp,&thresh_fp,&floor_fp,&descripts_fp,
			DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));
  floors = (int *) calloc(maxblocksize,sizeof(int));
  perm = (short *) calloc(maxblocksize,sizeof(short));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    potential_positions = 0;
    actual_positions = 0;
    skips = 0;

    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&perm[0],sizeof(short),width,mat_fp);
    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&floors[0],sizeof(int),width,floor_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    /* j is position along sequence */
    /* k is position along scoring matrix */
    /* i is a matrix pointer */
    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = -thresh;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + perm[k]]];
	  if (score < floors[k]) {
	    actual_positions += k+1;
	    potential_positions += width;
	    ++skips;
	    goto too_low;
	  }
	}
	score += thresh;
	actual_positions += width;
	potential_positions += width;
	if (Num_positives >= MAXPOSITIVES) {
	  printf("Error! Too many positives: Limit is %d\n",
		 MAXPOSITIVES);
	  exit(1);
	} else {
	  hit_copy(hit,sequence_char,j,width,length);
	  prob = interpolate_score(score,thresholds);
	  fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	  fread(&description[0],sizeof(char),descript_length,descripts_fp);
	  last_descript_ptr = descript_ptr + (long) descript_length;
	  Entries[Num_positives++] = 
	    Entry_new(description,descript_length,j,width,score,prob,info,hit);
	}
      too_low: ;
      }
    }
    printf("%d %f\n",width,
	   1.0 - (double) actual_positions/(double) potential_positions);
  }
  free(perm);
  free(floors);
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(floor_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

void
print_entries () {
  int i;
  
  for (i = 0; i < Num_positives; ++i) {
    printf("%4d. ",i+1);
    Entry_print(Entries[i]);
  }
}


/* Note: the following arguments are intended only for research
benchmarking, and are not intended for users to call:
   -F means use floor
   -B means do benchmark, floor vs standard
   -P means use ordered
   -B -P means do benchmark, ordered vs standard
*/

static void
parse_arguments (bool *benchmarkp, bool *floorp, bool *orderedp, bool *quietp,
		 int *cutoff_sig, float *threshold_info,
		 char **prefix, char **seqfile, int argc, char *argv[]) {
  ++argv; --argc;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if ((*argv)[1] == 's') {
	*cutoff_sig = atoi(argv[1]); argc -= 2; ++argv; ++argv;
	if (*cutoff_sig <= 0) {
	  fprintf(stderr,"Cutoff significance must be greater than 0\n");
	  exit(1);
	} else if (*cutoff_sig > MAXSIGNIF) {
	  fprintf(stderr,"Cutoff significance must be %d or less\n",MAXSIGNIF);
	  exit(1);
	}
      } else if ((*argv)[1] == 'B') {
	*benchmarkp = TRUE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'F') {
	*floorp = TRUE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'N') {
	*floorp = FALSE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'A') {
	*orderedp = FALSE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'P') {
	*orderedp = TRUE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'Q') {
	*quietp = TRUE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'i') {
	*threshold_info = atof(argv[1]); argc -= 2; ++argv; ++argv;
      } else if ((*argv)[1] == 'f') {
	++argv; *prefix = *argv; argc -= 2; ++argv;
      } else {
	printf("I don't understand.\n");
	exit(1);
      }
    } else {
      *seqfile = *argv; argc -= 1; argv += 1;
    }
  }
}

int 
main (int argc, char **argv) {
  bool benchmarkp = FALSE, floorp = TRUE, orderedp = TRUE, quietp = FALSE;
  int cutoff_sig = DEFAULT_SEARCH_THRESHOLD, length;
  float threshold_info = 0.0;
  char *seqfile = NULL, c;
  FILE *fp;
  int *sequence_int, i;
  char *sequence_char, *name, *prefix = "blocksplus";

  parse_arguments(&benchmarkp,&floorp,&orderedp,&quietp,
		  &cutoff_sig,&threshold_info,&prefix,&seqfile,argc,argv);
  if (seqfile == NULL) {
    fp = stdin;
  } else {
    if (!(fp = fopen(seqfile,"r"))) {
      fprintf(stderr,"Can't open sequence file %s\n",seqfile);
      exit(1);
    }
  }

  if (benchmarkp) {
    printf("blockwidth savings\n");
    while (!feof(fp) && !isspace(c = fgetc(fp))) {
      ungetc(c,fp);
      if (sequence_char = Sequence_read(fp,&name,&length)) {
	sequence_int = map_sequence(sequence_char,length);
	if (orderedp) {
	  score_sequence_bench_ordered(sequence_int,sequence_char,length,
				       cutoff_sig,prefix,threshold_info);
	} else {
	  score_sequence_bench(sequence_int,sequence_char,length,cutoff_sig,
			       prefix,threshold_info);
	}	  
	free(sequence_int);
	free(sequence_char);
	for (i = 0; i < Num_positives; ++i) {
	  Entry_free(&(Entries[i]));
	}
      }
    }
  } else if (orderedp) {
    while (!feof(fp) && !isspace(c = fgetc(fp))) {
      ungetc(c,fp);
      if (sequence_char = Sequence_read(fp,&name,&length)) {
	sequence_int = map_sequence(sequence_char,length);
	score_sequence_ordered(sequence_int,sequence_char,length,cutoff_sig,
			       prefix,threshold_info);
	free(sequence_int);
	free(sequence_char);
	qsort(Entries,Num_positives,sizeof(Entry_T),Entry_cmp);
	if (!quietp) {
	  if (name != NULL) {
	    printf("eMatrix search results for: %s\n\n",name);
	  }
	  print_entries();
	  printf("-------------------------------------------------\n\n");
	}
	for (i = 0; i < Num_positives; ++i) {
	  Entry_free(&(Entries[i]));
	}
      }
    }
  } else if (floorp) {
    while (!feof(fp) && !isspace(c = fgetc(fp))) {
      ungetc(c,fp);
      if (sequence_char = Sequence_read(fp,&name,&length)) {
	sequence_int = map_sequence(sequence_char,length);
	score_sequence_floor(sequence_int,sequence_char,length,cutoff_sig,
			       prefix,threshold_info);
	free(sequence_int);
	free(sequence_char);
	qsort(Entries,Num_positives,sizeof(Entry_T),Entry_cmp);
	if (!quietp) {
	  if (name != NULL) {
	    printf("eMatrix search results for: %s\n\n",name);
	  }
	  print_entries();
	  printf("-------------------------------------------------\n\n");
	}
	for (i = 0; i < Num_positives; ++i) {
	  Entry_free(&(Entries[i]));
	}
      }
    }
  } else {
    while (!feof(fp) && !isspace(c = fgetc(fp))) {
      ungetc(c,fp);
      if (sequence_char = Sequence_read(fp,&name,&length)) {
	sequence_int = map_sequence(sequence_char,length);
	score_sequence(sequence_int,sequence_char,length,cutoff_sig,
		       prefix,threshold_info);
	free(sequence_int);
	free(sequence_char);
	qsort(Entries,Num_positives,sizeof(Entry_T),Entry_cmp);
	if (!quietp) {
	  if (name != NULL) {
	    printf("eMatrix search results for: %s\n\n",name);
	  }
	  print_entries();
	  printf("-------------------------------------------------\n\n");
	  for (i = 0; i < Num_positives; ++i) {
	    Entry_free(&(Entries[i]));
	  }
	}
      }
    }
  }

  if (seqfile != NULL) {
    fclose(fp);
  }
  return 0;
}



